/*   FILE: qtconv.c -- 
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 20 July 2001
 *
 * Copyright (C) 2001 W. Michael Petullo <mike@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <quicktime/quicktime.h>
#include <getopt.h>
#include <string.h>

#define SAMPLES 4096		/* quicktime_encode_audio doesn't seem to like to */
			/* handle less than 4096 samples at a time when   */
			/* encoding Ogg Vorbis.  See SourceForge forum.   */

extern char *usage_qtconv;

/* ============================ settings_t ================================= */
typedef struct settings_t {
    char target_audio[5];
    char target_video[5];
    quicktime_t *input;
    quicktime_t *output;
    int quality;
    int cpus;
} settings_t;

/* ============================ parameters_t =============================== */
typedef struct parameters_t {
    char *copyright;
    char *name;
    char *info;
    int atracks;
    int vtracks;
    long *sample_rate;
    int *audio_bits;
    int *channels;
    int *width;
    int *height;
    int *depth;
    float *framerate;
} parameters_t;

/* ============================ usage () ==================================== */
void usage(const int exitcode, const char *error, const char *more)
{
    fprintf(stderr, "%s\n", usage_qtconv);
    if (error)
	fprintf(stderr, "%s: %s.\n\n", error, more);
    exit(exitcode);
}

/* ============================ parse_args () =============================== */
void parse_args(int argc, char *argv[], settings_t * s)
{
    char c;
    int opt_index = 0;
    struct option opts[] = {
	{"help", 0, 0, 'h'},
	{"audio", 1, 0, 'a'},
	{"video", 1, 0, 'v'},
	{"input", 1, 0, 'i'},
	{"output", 1, 0, 'o'},
	{"quality", 1, 0, 'q'},
	{"cpus", 1, 0, 'c'},
	{0, 0, 0, 0}
    };
    strcpy(s->target_audio, "twos");
    strcpy(s->target_video, "jpeg");
    s->input = NULL;
    s->output = NULL;
    s->quality = 0;
    s->cpus = 1;
    while ((c = getopt_long(argc, argv, "ha:v:i:o:q:c:", opts, &opt_index))
	   >= 0) {
	switch (c) {
	case 'h':
	    usage(EXIT_SUCCESS, NULL, NULL);
	case 'a':
	    strncpy(s->target_audio, optarg, sizeof(s->target_audio) - 1);
	    break;
	case 'v':
	    strncpy(s->target_video, optarg, sizeof(s->target_video) - 1);
	    break;
	case 'i':
	    s->input = quicktime_open(optarg, 1, 0);
	    break;
	case 'o':
	    s->output = quicktime_open(optarg, 0, 1);
	    break;
	case 'q':
	    s->quality = atoi(optarg);
	    break;
	case 'c':
	    s->cpus = atoi(optarg);
	    break;
	default:
	    fprintf(stderr, "%s\n", usage_qtconv);
	    exit(EXIT_FAILURE);
	}
    }
    if (!s->input)
	usage(EXIT_FAILURE, "qtconv", "no valid input path specified");
    if (!s->output)
	usage(EXIT_FAILURE, "qtconv", "no valid output path specified");
    if (s->quality > 0 && strcmp(s->target_video, "jpeg")
	&& strcmp(s->target_video, "mjpa"))
	usage(EXIT_FAILURE, "qtconv",
	      "JPEG option specified but target format not JPEG or MJPA");
    if (s->cpus > 1 && strcmp(s->target_video, "mjpa"))
	usage(EXIT_FAILURE, "qtconv",
	      "MJPA option specified but target format not MJPA");
}

/* ============================ set_parameters () ========================== */
int set_parameters(settings_t s, parameters_t * p)
{
    int i;
    p->copyright = quicktime_get_copyright(s.input);
    p->name = quicktime_get_name(s.input);
    p->info = quicktime_get_info(s.input);
    p->atracks = quicktime_audio_tracks(s.input);
    p->sample_rate = (long *) malloc(p->atracks * sizeof(long));
    p->audio_bits = (int *) malloc(p->atracks * sizeof(int));
    p->channels = (int *) malloc(p->atracks * sizeof(int));
    for (i = 0; i < p->atracks; i++) {
	p->sample_rate[i] = quicktime_sample_rate(s.input, i);
	p->audio_bits[i] = quicktime_audio_bits(s.input, i);
	p->channels[i] = quicktime_track_channels(s.input, i);
    }
    p->vtracks = quicktime_video_tracks(s.input);
    p->width = (int *) malloc(p->vtracks * sizeof(int));
    p->height = (int *) malloc(p->vtracks * sizeof(int));
    p->depth = (int *) malloc(p->vtracks * sizeof(int));
    p->framerate = (float *) malloc(p->vtracks * sizeof(float));
    for (i = 0; i < p->vtracks; i++) {
	p->width[i] = quicktime_video_width(s.input, i);
	p->height[i] = quicktime_video_height(s.input, i);
	p->depth[i] = quicktime_video_depth(s.input, i);
	p->framerate[i] = quicktime_frame_rate(s.input, i);
    }
    return 1;
}

/* ============================ convert_audio () =========================== */
int convert_audio(settings_t s, parameters_t p, float time)
{
    int i;
    static long audio_samples_written = 0;
    for (i = 0; i < p.atracks; i++) {
	int ch;
	int16_t **abuffer;
	long frame;
	long samples = time * p.sample_rate[i];
	abuffer = (int16_t **) malloc(p.channels[i] * sizeof(int16_t *));
	for (ch = 0; ch < p.channels[i]; ch++)
	    abuffer[ch] =
		(int16_t *) malloc(sizeof(int16_t) * SAMPLES *
				   (p.audio_bits[i] / 8));
	for (frame = 0; frame < samples; frame += SAMPLES) {
	    long n = frame + SAMPLES > samples ? samples - frame : SAMPLES;
	    for (ch = 0; ch < p.channels[i]; ch++) {
		quicktime_set_audio_position(s.input, audio_samples_written, 0); /* ch. not indep. */
		quicktime_decode_audio(s.input, abuffer[ch], NULL, n, ch);
	    }
	    quicktime_encode_audio(s.output, abuffer, NULL, n);
	    audio_samples_written += n;
	}
	for (ch = 0; ch < p.channels[i]; ch++)
	    free(abuffer[ch]);
	free(abuffer);
    }
    return 1;
}

/* ============================ convert_video () =========================== */
int convert_video(settings_t s, parameters_t p)
{
    int i;
    for (i = 0; i < p.vtracks; i++) {
	unsigned char **vbuffer;
	int j;
	vbuffer = (unsigned char **) malloc(p.height[i] * sizeof(char *));
	for (j = 0; j < p.height[i]; j++)
	    vbuffer[j] =
		(unsigned char *) malloc((p.width[i] * p.depth[i] / 8) *
					 sizeof(char));
	quicktime_decode_video(s.input, vbuffer, i);
	quicktime_encode_video(s.output, vbuffer, i);
	for (j = 0; j < p.height[i]; j++)
	    free(vbuffer[j]);
	free(vbuffer);
    }
    return 1;
}

/* ============================ free_parameters () ========================= */
void free_parameters(parameters_t p)
{
    free(p.sample_rate);
    free(p.audio_bits);
    free(p.channels);
    free(p.width);
    free(p.height);
    free(p.depth);
    free(p.framerate);
}

/* ========================== main () ======================================= */
int main(int argc, char *argv[], char *env[])
{
    long i, frames = 0;
    float audio_time = 0, video_time = 0;

    int vorbis_bitrate = 128000;
    int vorbis_min_bitrate = -1;
    int vorbis_max_bitrate = -1;

    int divx_bitrate = 1000000;
    int divx_rc_period = 50;
    int divx_rc_reaction_ratio = 45;
    int divx_rc_reaction_period = 10;
    int divx_max_key_interval = 45;
    int divx_max_quantizer = 31;
    int divx_min_quantizer = 1;
    int divx_quantizer = 10;
    int divx_quality = 5;
    int divx_fix_bitrate = 1;

    settings_t s;
    parameters_t p;
    parse_args(argc, argv, &s);
    set_parameters(s, &p);

    quicktime_set_copyright(s.output, p.copyright ? p.copyright : "");
    quicktime_set_name(s.output, p.name ? p.name : "");
    quicktime_set_info(s.output, p.info ? p.info : "");

    if (!strcmp(s.target_audio, QUICKTIME_VORBIS)) {
	printf("setting Ogg Vorbis specific parameters...\n");
	quicktime_set_parameter(s.output, "vorbis_bitrate",
				&vorbis_bitrate);
	quicktime_set_parameter(s.output, "vorbis_min_bitrate",
				&vorbis_min_bitrate);
	quicktime_set_parameter(s.output, "vorbis_max_bitrate",
				&vorbis_max_bitrate);
    }

    if (!strcmp(s.target_video, QUICKTIME_DIVX)) {
	printf("setting DivX specific parameters...\n");
	quicktime_set_parameter(s.output, "divx_bitrate", &divx_bitrate);
	quicktime_set_parameter(s.output, "divx_rc_period",
				&divx_rc_period);
	quicktime_set_parameter(s.output, "divx_rc_reaction_ratio",
				&divx_rc_reaction_ratio);
	quicktime_set_parameter(s.output, "divx_rc_reaction_period",
				&divx_rc_reaction_period);
	quicktime_set_parameter(s.output, "divx_max_key_interval",
				&divx_max_key_interval);
	quicktime_set_parameter(s.output, "divx_max_quantizer",
				&divx_max_quantizer);
	quicktime_set_parameter(s.output, "divx_min_quantizer",
				&divx_min_quantizer);
	quicktime_set_parameter(s.output, "divx_quantizer",
				&divx_quantizer);
	quicktime_set_parameter(s.output, "divx_quality", &divx_quality);
	quicktime_set_parameter(s.output, "divx_fix_bitrate",
				&divx_fix_bitrate);
    } else if (!(strcmp(s.target_video, "jpeg")
	  && strcmp(s.target_video, "mjpa"))) {
	quicktime_set_jpeg(s.output, s.quality > 0 ? s.quality : 75, 1);
	printf("set JPEG quality to %d\n", s.quality > 0 ? s.quality : 75);
    }

    quicktime_set_cpus(s.output, s.cpus);
    quicktime_set_audio(s.output, p.channels[0], p.sample_rate[0],
			p.audio_bits[0], s.target_audio);
    quicktime_set_video(s.output, p.vtracks, p.width[0], p.height[0],
			p.framerate[0], s.target_video);

    for (i = 0; i < p.atracks; i++) {
	if (!quicktime_supported_audio(s.input, i)) {
	    fprintf(stderr, "qtconf: unsupported input audio\n");
	    exit(1);
	}
	if (!quicktime_supported_audio(s.output, i)) {
	    fprintf(stderr, "qtconf: unsupported output audio\n");
	    exit(1);
	}
    }

    for (i = 0; i < p.vtracks; i++) {
	quicktime_set_depth(s.output, p.depth[i], i);
	if (!quicktime_supported_video(s.input, i)) {
	    fprintf(stderr, "qtconf: unsupported input video\n");
	    exit(1);
	}
	if (!quicktime_supported_video(s.output, i)) {
	    fprintf(stderr, "qtconf: unsupported output video\n");
	    exit(1);
	}
    }

    while (frames < quicktime_video_length(s.input, 0)) {
	if (audio_time < video_time) {
	    printf("writing audio frames...");
	    convert_audio(s, p, video_time - audio_time);
	    audio_time = video_time;
	    printf(" done\n");
	} else {
	    printf("writing video frame %ld of %ld...", frames,
		   quicktime_video_length(s.input, 0));
	    convert_video(s, p);
	    video_time += (1 / p.framerate[0]);
	    frames++;
	    printf(" done\n");
	}
    }

    printf("closing output file...\n");
    quicktime_close(s.output);
    printf("closing input file...\n");
    quicktime_close(s.input);
    printf("done\n");
    free_parameters(p);
    exit(EXIT_SUCCESS);
}
