/*   FILE: qtsg.c -- 
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 23 July 2001
 *
 * Copyright (C) 2001 W. Michael Petullo <mike@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include <quicktime/quicktime.h>
#include <jpeglib.h>

extern char *usage_qtsg;

/* ============================ settings_t ================================= */
typedef struct settings_t {
    char target_audio[5];
    char *output;
    char *audio;
    int length;
    float frame_rate;
} settings_t;

/* ============================ usage () ==================================== */
void usage(const int exitcode, const char *error, const char *more)
{
    fprintf(stderr, "%s\n", usage_qtsg);
    if (error)
	fprintf(stderr, "%s: %s.\n\n", error, more);
    exit(exitcode);
}

/* ============================ parse_args () =============================== */
void parse_args(int argc, char *argv[], settings_t * settings)
{
    char c;
    int opt_index = 0;
    struct option opts[] = {
	{"help", 0, 0, 'h'},
	{"output", 1, 0, 'i'},
	{"audio", 1, 0, 'a'},
	{"length", 1, 0, 'l'},
	{"acodec", 1, 0, 'c'},
	{0, 0, 0, 0}
    };
    settings->output = NULL;
    settings->audio = NULL;
    settings->length = 0;
    settings->frame_rate = 1;
    strcpy(settings->target_audio, "twos");
    while ((c = getopt_long(argc, argv, "ho:a:l:c:r:", opts, &opt_index))
	   >= 0) {
	switch (c) {
	case 'h':
	    usage(EXIT_SUCCESS, NULL, NULL);
	case 'o':
	    settings->output = optarg;
	    break;
	case 'a':
	    settings->audio = optarg;
	    break;
	case 'l':
	    settings->length = atoi(optarg);
	    break;
	case 'c':
	    strncpy (settings->target_audio, optarg, 4);
	    break;
	case 'r':
	    settings->frame_rate = atof(optarg);
	    break;
	default:
	    usage(EXIT_FAILURE, NULL, NULL);
	}
    }
    if (!settings->output)
	usage(EXIT_FAILURE, "qtsg", "no valid output path specified");
    if (settings->audio && settings->length > 0)
	usage(EXIT_FAILURE, "qtsg",
	      "both audio input and length specified");
}

/* ============================ add_image () =============================== */
void add_image(quicktime_t * output, FILE * image)
{
    struct stat stat_buf;
    unsigned char *buf;
    fstat(fileno(image), &stat_buf);
    buf = (char *) malloc(stat_buf.st_size * sizeof(char));
    fread(buf, sizeof(char), stat_buf.st_size, image);
    quicktime_write_frame(output, buf, stat_buf.st_size, 0);
    free(buf);
}

/* ============================ add_video () =============================== */
void add_video(quicktime_t * output, quicktime_t * input)
{
    unsigned char **buf;
    int width, height, i, depth;
    long length;
    length = quicktime_video_length(input, 0);
    width = quicktime_video_width(input, 0);
    height = quicktime_video_height(input, 0);
    depth = quicktime_video_depth(input, 0);
    buf = (unsigned char **) malloc(height * sizeof(char *));
    for (i = 0; i < height; i++)
	buf[i] =
	    (unsigned char *) malloc(width * (depth / 8) * sizeof(char));
    quicktime_decode_video(input, buf, 0);
    quicktime_encode_video(output, buf, 0);
    for (i = 0; i < height; i++)
	free(buf[i]);
    free(buf);
}

/* ============================ add_audio () =============================== */
void add_audio(quicktime_t * qt, OggVorbis_File * audio, int ch, int word,
	       float time)
{
    int sect;
    char buf[BUFSIZ];
    int16_t **chans;
    long bytes, written = 0;
    long samples_goal = time * quicktime_sample_rate(qt, 0);
    while (((written < samples_goal * word * ch)
	    && (bytes = ov_read(audio, buf, BUFSIZ, 1, word, 1, &sect)))) {
	int i, j;
	if (bytes < 0)
	    fprintf(stderr, "error reading audio input\n");
	else {
	    /* Convert to QT4L-centric data first. */
	    int step = ch * word;
	    long samples = bytes / step;
	    chans = (int16_t **) malloc(sizeof(int16_t *) * ch);
	    for (i = 0; i < ch; i++)
		chans[i] = (int16_t *) malloc(sizeof(int16_t) * bytes / 2);

	    /* FIXME Seems to play slow. */
	    for (i = 0; i < ch; i++) {
		for (j = 0; j < samples; j++) {
		    chans[i][j] = (((int16_t) buf[j * step + i * 2]) << 8) +
			((unsigned char) buf[j * step + i * 2 + 1]);
		}
	    }

	    quicktime_encode_audio(qt, chans, NULL, bytes / ch / word);
	    for (i = 0; i < ch; i++)
		free(chans[i]);
	    free(chans);
	    written += bytes;
	}
    }
}

/* ============================ set_width_and_height () ==================== */
void set_width_and_height(char *path, int *width, int *height)
{
    FILE *i;
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
    if (!(i = fopen(path, "r"))) {
	perror("error opening image file");
	exit(EXIT_FAILURE);
    }
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, i);
    jpeg_read_header(&cinfo, TRUE);
    *width = cinfo.image_width;
    *height = cinfo.image_height;
    jpeg_destroy_decompress(&cinfo);
    fclose(i);
}

/* ============================ main () ===================================== */
int main(int argc, char *argv[], char *env[])
{
    FILE *a = NULL, *image;
    vorbis_info *vi;
    int word = 2;		// Number of bytes in a sample.  2 means 16 bit audio.
    long audio_length = 0;
    settings_t settings;
    quicktime_t *output;
    OggVorbis_File audio;
    int i, num_inputs;
    int video_width, video_height;

    parse_args(argc, argv, &settings);
    output = quicktime_open(settings.output, 0, 1);
    if (settings.audio) {
	if (!(a = fopen(settings.audio, "r"))) {
	    perror("error opening audio file");
	    exit(EXIT_FAILURE);
	}
	if (ov_open(a, &audio, NULL, 0)) {
	    fprintf(stderr, "error opening audio file\n");
	    exit(EXIT_FAILURE);
	}
	audio_length = ov_time_total(&audio, -1);
	vi = ov_info(&audio, -1);
	quicktime_set_audio(output, vi->channels, vi->rate, 8 * word,
			    settings.target_audio);
	printf("opened audio file.  Length is %ld sec.\n", audio_length);
    } else if (settings.length > 0)
	audio_length = settings.length;
    set_width_and_height(argv[optind], &video_width, &video_height);
    num_inputs = argc - optind;
    printf("With %d images, each image will appear for %ld seconds.\n",
	   num_inputs, audio_length / num_inputs);
    quicktime_set_video(output, 1, video_width, video_height, settings.frame_rate,
			QUICKTIME_JPEG);
    for (i = optind; argv[i]; i++) {
	long f = 0, frames;
	float audio_time = 0, video_time = 0;
	quicktime_t *video;
	int image_width, image_height, is_video = 0;
	printf("Adding %s...\n", argv[i]);
	set_width_and_height(argv[i], &image_width, &image_height);
	if (!(image_width == video_width && image_height == video_height)) {
	    fprintf(stderr,
		    "qtsg: inconsistent image resolution -- try qtscaledsg\n");
	    exit(EXIT_FAILURE);
	}

	if (quicktime_check_sig(argv[i])) {
	    video = quicktime_open(argv[i], 1, 0);
	    frames = quicktime_video_length(video, 0);
	    is_video = 1;
	} else {
	    if (!(image = fopen(argv[i], "r"))) {
		perror("error opening image file");
		exit(EXIT_FAILURE);
	    }
	    frames = audio_length ? audio_length / num_inputs : 1;
	}
	printf("%ld\n", frames);
	while (f < frames) {
	    if (audio_time < video_time && settings.audio) {
		printf("writing audio frames...");
		add_audio(output, &audio, vi->channels, word,
			  video_time - audio_time);
		audio_time = video_time;
		printf(" done\n");
	    } else {
		printf("writing video frame %ld of %ld...", f, frames);
		if (is_video)
		    add_video(output, video);
		else
		    add_image(output, image);
		video_time += (1 / quicktime_frame_rate(output, 0));
		f++;
		printf(" done\n");
	    }
	}
	if (is_video)
	    quicktime_close(video);
	else
	    fclose(image);
    }
    quicktime_close(output);
    if (a)
        fclose(a);
    exit(EXIT_SUCCESS);
}
