/*   FILE: qtsplit.c -- 
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 20 July 2001
 *
 * Copyright (C) 2001 W. Michael Petullo <mike@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <quicktime/quicktime.h>
#include <getopt.h>
#include <string.h>
#include <limits.h>

extern char *usage_qtsplit;

/* ============================ settings_t ================================= */
typedef struct settings_t {
    quicktime_t *input;
    char output[PATH_MAX + 1];
    long chunk_size;
} settings_t;

/* ============================ parameters_t =============================== */
typedef struct parameters_t {
    char *copyright;
    char *name;
    char *info;
    int atracks;
    int vtracks;
    long *sample_rate;
    int *audio_bits;
    int *channels;
    int *width;
    int *height;
    int *depth;
    float *framerate;
    char **v_compressor;
    char **a_compressor;
} parameters_t;

/* ============================ usage () ==================================== */
void usage(const int exitcode, const char *error, const char *more)
{
    fprintf(stderr, "%s\n", usage_qtsplit);
    if (error)
	fprintf(stderr, "%s: %s.\n\n", error, more);
    exit(exitcode);
}

/* ============================ parse_args () =============================== */
void parse_args(int argc, char *argv[], settings_t * s)
{
    char c;
    int opt_index = 0;
    struct option opts[] = {
	{"help", 0, 0, 'h'},
	{"input", 1, 0, 'i'},
	{"output", 1, 0, 'o'},
	{"chunk_size", 1, 0, 's'},
	{0, 0, 0, 0}
    };
    s->input = NULL;
    s->chunk_size = 630;
    strcpy(s->output, "");
    while ((c = getopt_long(argc, argv, "hi:o:s:", opts, &opt_index))
	   >= 0) {
	switch (c) {
	case 'h':
	    usage(EXIT_SUCCESS, NULL, NULL);
	case 'i':
	    s->input = quicktime_open(optarg, 1, 0);
	    break;
	case 'o':
	    strncpy(s->output, optarg, PATH_MAX);
	    s->output[PATH_MAX] = 0x00;
	    break;
	case 's':
	    s->chunk_size = atoi(optarg);
	    break;
	default:
	    fprintf(stderr, "%s\n", usage_qtsplit);
	    exit(EXIT_FAILURE);
	}
    }
    if (!s->input)
	usage(EXIT_FAILURE, "qtsplit", "no valid input path specified");
    if (!strlen(s->output))
	usage(EXIT_FAILURE, "qtsplit", "no valid output path specified");
    if (!strstr(s->output, "%d"))
	usage(EXIT_FAILURE, "qtsplit", "invalid output path: no %d");
}

/* ============================ set_parameters () ========================== */
int set_parameters(settings_t s, parameters_t * p)
{
    int i;
    p->copyright = quicktime_get_copyright(s.input);
    p->name = quicktime_get_name(s.input);
    p->info = quicktime_get_info(s.input);
    p->atracks = quicktime_audio_tracks(s.input);
    p->sample_rate = (long *) malloc(p->atracks * sizeof(long));
    p->audio_bits = (int *) malloc(p->atracks * sizeof(int));
    p->channels = (int *) malloc(p->atracks * sizeof(int));
    for (i = 0; i < p->atracks; i++) {
	p->sample_rate[i] = quicktime_sample_rate(s.input, i);
	p->audio_bits[i] = quicktime_audio_bits(s.input, i);
	p->channels[i] = quicktime_track_channels(s.input, i);
    }
    p->vtracks = quicktime_video_tracks(s.input);
    p->width = (int *) malloc(p->vtracks * sizeof(int));
    p->height = (int *) malloc(p->vtracks * sizeof(int));
    p->depth = (int *) malloc(p->vtracks * sizeof(int));
    p->framerate = (float *) malloc(p->vtracks * sizeof(float));
    p->v_compressor = (char **) malloc(p->vtracks * sizeof(char *));
    p->a_compressor = (char **) malloc(p->vtracks * sizeof(char *));
    for (i = 0; i < p->vtracks; i++) {
	p->width[i] = quicktime_video_width(s.input, i);
	p->height[i] = quicktime_video_height(s.input, i);
	p->depth[i] = quicktime_video_depth(s.input, i);
	p->framerate[i] = quicktime_frame_rate(s.input, i);
	p->v_compressor[i] = quicktime_video_compressor(s.input, i);
	p->a_compressor[i] = quicktime_audio_compressor(s.input, i);
    }
    return 1;
}

/* ============================ copy_audio () ============================== */
long copy_audio(settings_t s, parameters_t p, quicktime_t * q)
{
    int i;
    long bytes = 0;
    for (i = 0; i < p.atracks; i++) {
	long samples =
	    (long int) ((1 / p.framerate[i]) * p.sample_rate[i]);
	long size =
	    (long int) samples * (p.audio_bits[i] / 8) * p.channels[i] *
	    sizeof(char);
	char *buffer = (char *) malloc(size);
	bytes += quicktime_read_audio(s.input, buffer, samples, i);
	quicktime_write_audio(q, buffer, samples, i);
	free(buffer);
    }
    return bytes;
}

/* ============================ copy_frame () ============================== */
long copy_frame(settings_t s, parameters_t p, quicktime_t * q)
{
    int i;
    long bytes = 0;
    for (i = 0; i < p.vtracks; i++) {
	long frame_size;
	char *buffer =
	    (char *) malloc(p.width[i] * p.height[i] * p.depth[i] *
			    sizeof(char));
	bytes += frame_size = quicktime_read_frame(s.input, buffer, i);
	quicktime_write_frame(q, buffer, frame_size, i);
	free(buffer);
    }
    return bytes;
}

/* ============================ free_parameters () ========================= */
void free_parameters(parameters_t p)
{
    free(p.sample_rate);
    free(p.audio_bits);
    free(p.channels);
    free(p.width);
    free(p.height);
    free(p.depth);
    free(p.framerate);
    free(p.v_compressor);
    free(p.a_compressor);
}

/* ============================ write_chunck () ============================ */ 
int write_chunk(long vid_len, settings_t s, parameters_t p, quicktime_t *o)
/* Returns 1 when last frame of video has been written, otherwise 0. */
{
    static long num_frames = 0;
    static int chunk = 0;
    long bytes = 0;
    while (bytes < s.chunk_size * 1024 * 1024) {
        bytes += copy_frame(s, p, o);
        bytes += copy_audio(s, p, o);
        if (++num_frames == vid_len) {
            return 1;
            printf("chunk %d (%ld megabytes) done\n", chunk, bytes / 1024 / 1024);
	}
    }
    printf("chunk %d (%ld megabytes) done\n", chunk, bytes / 1024 / 1024);
    chunk++;
    return 0;
}

/* ============================ main () ==================================== */
int main(int argc, char *argv[], char *env[])
{
    int chunk;
    settings_t s;
    parameters_t p;
    char path[PATH_MAX + 1];
    int done = 0;
    long video_length;
    parse_args(argc, argv, &s);
    set_parameters(s, &p);
    video_length = quicktime_video_length(s.input, 0);
    for (chunk = 0; !done; chunk++) {
	quicktime_t *output;
	sprintf(path, s.output, chunk);
	output = quicktime_open(path, 0, 1);
	quicktime_set_copyright(output, p.copyright ? p.copyright : "");
	quicktime_set_name(output, p.name ? p.name : "");
	quicktime_set_info(output, p.info ? p.info : "");
	quicktime_set_video(output, p.vtracks, p.width[0], p.height[0],
			    p.framerate[0], p.v_compressor[0]);
	quicktime_set_audio(output, p.channels[0], p.sample_rate[0],
			    p.audio_bits[0], p.a_compressor[0]);
	quicktime_set_depth(output, p.depth[chunk], chunk);
	done = write_chunk(video_length, s, p, output);
	quicktime_close(output);
    }
    free_parameters(p);
    quicktime_close(s.input);
    exit(EXIT_SUCCESS);
}
